import pytest
from loguru import logger
import pathlib
import pickle
from decimal import Decimal
from typing import Dict, List
import datetime as dt
import shioaji as sj
from shioaji import Exchange

from coverbot.cover import CoverBot

from .quote import QuoteSTKv1


contract_filename = f"{pathlib.Path(__file__).parent.absolute()}/data/contracts.pkl"


@pytest.fixture
def deal_data() -> List[Dict]:
    deal_filename = f"{pathlib.Path(__file__).parent.absolute()}/data/_deals.pkl"
    with open(deal_filename, "rb") as f:
        test_data = pickle.load(f)
    return test_data


@pytest.fixture
def quote_data():
    quote = QuoteSTKv1(
        code="1795",
        datetime=dt.datetime(2022, 7, 1, 10, 43, 15, 840092),
        open=Decimal("471.5"),
        avg_price=Decimal("467.9"),
        close=Decimal("461"),
        high=Decimal("474"),
        low=Decimal("461"),
        amount=Decimal("9220000"),
        total_amount=Decimal("11843696000"),
        volume=1,
        total_volume=25312,
        tick_type=2,
        chg_type=4,
        price_chg=Decimal("-15"),
        pct_chg=Decimal("-3.15"),
        bid_side_total_vol=9350,
        ask_side_total_vol=15962,
        bid_side_total_cnt=2730,
        ask_side_total_cnt=2848,
        closing_oddlot_shares=0,
        closing_oddlot_close=Decimal("0.0"),
        closing_oddlot_amount=Decimal("0"),
        closing_oddlot_bid_price=Decimal("0.0"),
        closing_oddlot_ask_price=Decimal("0.0"),
        fixed_trade_vol=0,
        fixed_trade_amount=Decimal("0"),
        bid_price=[
            Decimal("461"),
            Decimal("460.5"),
            Decimal("460"),
            Decimal("459.5"),
            Decimal("459"),
        ],
        bid_volume=[201, 141, 994, 63, 132],
        diff_bid_vol=[0, 1, 0, 0, 0],
        ask_price=[
            Decimal("461.5"),
            Decimal("462"),
            Decimal("462.5"),
            Decimal("463"),
            Decimal("463.5"),
        ],
        ask_volume=[123, 101, 103, 139, 95],
        diff_ask_vol=[0, 0, 0, 0, 0],
        avail_borrowing=9579699,
        suspend=0,
        simtrade=0,
    )
    exchange = Exchange("TSE")
    return exchange, quote


@pytest.fixture
def mock_api() -> sj.Shioaji:
    api: sj.Shioaji = sj.Shioaji()
    with open(contract_filename, "rb") as f:
        api.Contracts = pickle.load(f)
    return api


@pytest.fixture
def cover_bot(mock_api: sj.Shioaji) -> CoverBot:
    sut = CoverBot(mock_api)
    return sut



from loguru import logger
from typing import Dict, List, Optional
import math
import polars as pl
from dataclasses import dataclass, field, fields
from shioaji.constant import OrderState, Action, StockOrderCond
from shioaji.contracts import Contract
from shioaji import Exchange, QuoteSTKv1


@dataclass
class TFTDeal:
    code: str
    action: Action
    price: float
    quantity: int
    order_cond: StockOrderCond
    web_id: str
    custom_field: str
    time: str

    def __init__(self, **kwargs):
        names = set([f.name for f in fields(self)])
        for k, v in kwargs.items():
            if k in names:
                setattr(self, k, v)


class Deal:
    def __init__(self, contract: Contract):
        self.contract: Contract = contract
        self.is_cover: bool = False
        self.buy_quantity: int = 0
        self.sell_quantity: int = 0
        self.amount: float = 0
        self.stop_price: float = 0.0
        self.first_action: Optional[Action] = None
        self.entry_price: float = 0.0
        self.stop_loss_pct: float = 0.03
        self.buy_collection: List[TFTDeal] = []
        self.sell_collection: List[TFTDeal] = []

    @property
    def quantity(self):
        return self.buy_quantity - self.sell_quantity

    # def __repr__(self) -> str:
    #     return f"Code : {self.contract}\n quantity:{self.quantity} buy: {len(self.buy_collection)} sell: {len(self.sell_collection)}"
    def to_dict(self):
        d = {
            "contract": f"{self.contract.code} {self.contract.name}",
            "action": self.first_action,
            "quantity": self.quantity,
            "amount": self.amount,
            "entry_price": self.entry_price,
            "stop_price": self.stop_price,
            "buy_records": len(self.buy_collection),
            "sell_records": len(self.sell_collection),
        }
        return d

    def df(self):
        df = (
            pl.DataFrame(data=[d for d in self.buy_collection + self.sell_collection])
            .select(pl.col(["code", "name", "action", "price", "quantity", "ordno"]))
            .with_columns(
                [(pl.col("price") * pl.col("quantity") * 1000).alias("amount")]
            )
            .groupby(
                [pl.col("code"), pl.col("name"), pl.col("ordno"), pl.col("action")]
            )
            .agg([pl.col("quantity").sum(), pl.col("amount").sum()])
            .with_columns(
                [
                    (pl.col("amount") * 0.001425 * 0.2).apply(math.floor).alias("fee"),
                    (
                        pl.when(pl.col("action") == "Sell")
                        .then((pl.col("amount") * 0.0015).apply(math.floor))
                        .otherwise(0)
                        .alias("tax")
                    ),
                ]
            )
        )

    def apply_quote(self, exchange: Exchange, quote: QuoteSTKv1):
        if self.is_cover == True:
            return "is recoverd"
        if quote.simtrade == 0 and quote.volume > 0:
            if self.first_action == Action.Sell and quote.close >= self.stop_price:
                logger.info(
                    f"[{self.contract.code} {self.contract.name}] 均價 {self.entry_price} {self.quantity}張  賣單停損買回 現價 {quote.close} >= 停損價 {self.stop_price}"
                )
                self.is_cover = True
                return "place order sell"
            elif self.first_action == Action.Buy:
                logger.info(
                    f"[{self.contract.code} {self.contract.name}] 均價 {self.entry_price} {self.quantity}張  買單停損賣出 現價 {quote.close} <= 停損價 {self.stop_price}"
                )
                self.is_cover = True
                return "place order buy"
        return "nothing"

    def apply(self, tftdeal: Dict):
        deal = TFTDeal(**tftdeal)
        amount = deal.quantity * deal.price * 1000
        if not self.first_action:
            self.first_action = deal.action

        if deal.action == Action.Buy:
            self.buy_quantity += deal.quantity
            self.buy_collection.append(deal)

        elif deal.action == Action.Sell:
            self.sell_quantity += deal.quantity
            self.sell_collection.append(deal)

        if self.first_action == deal.action:
            self.amount += amount
            quantity = (
                self.buy_quantity
                if self.first_action == Action.Buy
                else self.sell_quantity
            )
            self.entry_price = self.amount / quantity / 1000.0
            stop_dvalue = self.entry_price * self.stop_loss_pct
            if self.first_action == Action.Buy:
                self.stop_price = self.entry_price - stop_dvalue
            elif self.first_action == Action.Sell:
                self.stop_price = self.entry_price + stop_dvalue



from loguru import logger
from typing import Dict, List, Any, Union
from dataclasses import dataclass, field, fields
import shioaji as sj
from shioaji.contracts import Contract

from shioaji.constant import OrderState, Action, StockOrderCond, QuoteType, QuoteVersion
from .deal import TFTDeal, Deal
from shioaji import QuoteSTKv1, Exchange


class CoverBot:
    def __init__(self, api: sj.Shioaji):
        self.api: sj.Shioaji = api
        self.api.set_order_callback(self.order_handler)
        self.api.quote.set_on_quote_stk_v1_callback(self.quote_handler)
        self.deals: Dict[str, Deal] = {}

    def set_stop_loss_pct(self, code: str, value: float):
        contract = self.api.Contracts.Stocks[code]
        deal = self.deals[code] = self.deals.get(code, Deal(contract))
        deal.stop_loss_pct = value

    def show(self):
        return [v.to_dict() for _, v in self.deals.items()]

    def deal_action(self, tftdeal: Dict[str, Any]) -> None:
        code = tftdeal.get("code", "")
        contract = self.api.Contracts.Stocks[code]
        if not contract:
            logger.warning(f"[{code}] not exist.")
        else:
            deal = self.deals[code] = self.deals.get(code, Deal(contract))
            deal.apply(tftdeal)

    def subscribe_quote(self, code: str):
        self.api.quote.subscribe(
            self.api.Contracts.Stocks[code],
            quote_type=QuoteType.Quote,
            version=QuoteVersion.v1,
        )

    def order_handler(self, order_state: OrderState, msg: Dict) -> None:
        if order_state == OrderState.TFTDeal:
            if msg["code"] not in self.deals.keys():
                self.subscribe_quote(msg["code"])
            self.deal_action(msg)
        elif order_state == OrderState.TFTOrder:
            if msg["contract"]["code"] not in self.deals.keys():
                self.subscribe_quote(msg["code"])

    def quote_handler(self, exchange: Exchange, quote: QuoteSTKv1):
        deal = self.deals.get(quote.code)
        if deal:
            result = deal.apply_quote(exchange, quote)
            logger.info(result)
            return result
        return None



from dataclasses import dataclass
import datetime as dt
from decimal import Decimal
from typing import List


@dataclass
class QuoteSTKv1:
    """Quote Stock v1
    Attributes:
        code (str): code
        datetime (datetime.datetime): datetime
        open (Decimal): open
        avg_price (Decimal): average price
        close (Decimal): deal price
        high (Decimal): high since market open
        low (Decimal): low since market open
        amount (Decimal): amount (NTD)
        total_amount (Decimal): total amount (NTD)
        volume (int): volume (K shares)
        total_volume (int): total volume (K shares)
        tick_type (int): tick type (內外盤別)
            {1: buy deal, 2: sell deal, 0: can't judge}
        chg_type (int): (漲跌註記)
            {1: limit up, 2: up, 3: unchanged, 4: down, 5: limit down}
        price_chg (Decimal): price change
        pct_chg (Decimal): percentage change (%)
        bid_side_total_vol(int): total bid deal volume (K shares, 買盤成交總量)
        ask_side_total_vol (int): total ask deal volume (K shares, 賣盤成交總量)
        bid_side_total_cnt (int): total number of buy deal (買盤成交筆數)
        ask_side_total_cnt (int): total number of sell deal (賣盤成交筆數)
        closing_oddlot_shares (int): (share, 盤後零股成交股數)
        closing_oddlot_close (Decimal): closing oddlot close
        closing_oddlot_amount (Decimal): closing oddlot amount
        closing_oddlot_bid_price (Decimal): closing oddlot bid price
        closing_oddlot_ask_price (Decimal): closing oddlot ask price
        fixed_trade_vol (int): fixed trade volume (K shares, 定盤成交量)
        fixed_trade_amount (Decimal): fixed trade amount
        bid_price (list of Decimal): bid price
        bid_volume (list of int): bid volume (lot), (張)
        diff_bid_vol (list of int): (lot), (張, 買價增減量)
        ask_price (list of Decimal): ask price
        ask_volume (list of int): ask volume (lot), (張)
        diff_ask_vol (list of int): (lot), (張, 賣價增減量)
        avail_borrowing (int): avail borrowing
        suspend (bool): suspend (暫停交易)
        simtrade (bool): simulated trade (試撮)
    """

    code: str
    datetime: dt.datetime
    open: Decimal
    avg_price: Decimal
    close: Decimal
    high: Decimal
    low: Decimal
    amount: Decimal
    total_amount: Decimal
    volume: int
    total_volume: int
    tick_type: int
    chg_type: int
    price_chg: Decimal
    pct_chg: Decimal
    bid_side_total_vol: int
    ask_side_total_vol: int
    bid_side_total_cnt: int
    ask_side_total_cnt: int
    closing_oddlot_shares: int
    closing_oddlot_close: Decimal
    closing_oddlot_amount: Decimal
    closing_oddlot_bid_price: Decimal
    closing_oddlot_ask_price: Decimal
    fixed_trade_vol: int
    fixed_trade_amount: Decimal
    bid_price: List[Decimal]
    bid_volume: List[int]
    diff_bid_vol: List[int]
    ask_price: List[Decimal]
    ask_volume: List[int]
    diff_ask_vol: List[int]
    avail_borrowing: int
    suspend: int
    simtrade: int



import pytest
from loguru import logger
import shioaji as sj
from shioaji import Exchange
from shioaji.constant import OrderState, StockOrderCond, Action
from typing import Tuple

# from coverbot.deal import TFTDeal
from coverbot.cover import CoverBot, TFTDeal
from .quote import QuoteSTKv1

deal_buy = {
    "trade_id": "eb6055e4",
    "seqno": "425813",
    "ordno": "IE324",
    "exchange_seq": "430956",
    "broker_id": "9A9S",
    "account_id": "0291589",
    "action": "Buy",
    "code": "5871",
    "order_cond": "Cash",
    "order_lot": "Common",
    "price": 191.5,
    "quantity": 1,
    "web_id": "137",
    "custom_field": "",
    "ts": 1668482197,
    "name": "中租-KY",
    "time": "2022-11-15 11:16:37",
}

deal_sell = {
    "trade_id": "eb6055e4",
    "seqno": "425813",
    "ordno": "IE324",
    "exchange_seq": "430956",
    "broker_id": "9A9S",
    "account_id": "0291589",
    "action": "Sell",
    "code": "5871",
    "order_cond": "Cash",
    "order_lot": "Common",
    "price": 191.5,
    "quantity": 1,
    "web_id": "137",
    "custom_field": "",
    "ts": 1668482197,
    "name": "中租-KY",
    "time": "2022-11-15 11:16:37",
}


def test_sell_deal_action(cover_bot: CoverBot):
    cover_bot.deal_action(deal_sell)
    cover_bot.show()
    assert cover_bot.deals[deal_sell["code"]].quantity == -1


def test_deal_action(cover_bot: CoverBot):
    assert cover_bot.deals.get(deal_buy["code"], None) == None
    cover_bot.deal_action(deal_buy)
    assert cover_bot.deals[deal_buy["code"]].quantity == 1
    cover_bot.deal_action(deal_sell)
    assert cover_bot.deals[deal_buy["code"]].quantity == 0
    cover_bot.deal_action(deal_sell)
    assert cover_bot.deals[deal_buy["code"]].quantity == -1


def test_deal_date(cover_bot: CoverBot):
    cover_bot.order_handler(OrderState.TFTDeal, deal_buy)
    assert cover_bot.deals[deal_buy["code"]].quantity == 1


def test_show(cover_bot: CoverBot, deal_data):
    for deal in deal_data:
        if deal["action"] == "Sell":
            cover_bot.order_handler(OrderState.TFTDeal, deal)
    logger.info(cover_bot.show())


def test_non_deal_quote_should_be_None(
    cover_bot: CoverBot, quote_data: Tuple[Exchange, QuoteSTKv1]
):
    exchange, quote = quote_data
    result = cover_bot.quote_handler(exchange=exchange, quote=quote)  # type: ignore
    assert result == None


def test_coverbot_set_stop_loss_pct(cover_bot: CoverBot):
    cover_bot.deal_action(deal_sell)
    code = deal_sell["code"]
    deal = cover_bot.deals[code]
    assert deal.stop_loss_pct == 0.03
    cover_bot.set_stop_loss_pct(code, 0.05)
    assert deal.stop_loss_pct == 0.05


def test_in_deal_quote_should_be_(
    cover_bot: CoverBot, quote_data: Tuple[Exchange, QuoteSTKv1]
):
    cover_bot.deal_action(deal_sell)
    exchange, quote = quote_data
    quote.code = "5871"
    result = cover_bot.quote_handler(exchange=exchange, quote=quote)  # type: ignore
    assert result == "place order sell"
